/*
 * SPDX-FileCopyrightText: 2010~2020 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#ifndef __FCITX_IM_CONTEXT_H_
#define __FCITX_IM_CONTEXT_H_

#include <gtk/gtk.h>

G_BEGIN_DECLS

GType fcitx5_im_context_get_type(void);
void fcitx5_im_context_register_type(GTypeModule *type_module);

G_END_DECLS
#endif
// kate: indent-mode cstyle; space-indent on; indent-width 0;
